# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""
Integer compositions
====================

An integer composition of a positive integer :math:`n` is a list of positive
integers :math:`k_1, k_2, \ldots, k_r` such that :math:`\sum_i k_i = n`. Using
the value of an integer as its size, we can identify the class of positive
integers with the class of non-empty sequences of atoms ``Seq(z, geq=1)`` by
identifying ``n`` with the unique list of length ``n``:

>>> from usainboltz import *
>>> from usainboltz.generator import seed
>>> seed(0xDEADBEEF)  # For reproducibility

>>> z = Atom()
>>> Nat = RuleName("Nat")
>>> grammar = Grammar({Nat: Seq(z, geq=1)})

Thus, integer partitions are non-empty sequences of integers:

>>> C = RuleName("C")
>>> grammar.add_rule(C, Seq(Nat, geq=1))
>>> grammar
C -> Seq(Nat, 1 <= .)
Nat -> Seq(z, 1 <= .)

Since the generating function of compositions is rational, one cannot use a
singular sampler. But instead, one can ask the generator to tune its oracle to
target a specific size (in expectation):

>>> generator = Generator(
...     grammar, C, expectations={z: 10}
... )

Finally, we define an appropriate builder that apply the bijection between lists
and integers:

>>> def build_nat(l):
...     return len(l)
>>> generator.set_builder(Nat, build_nat)

>>> from sage.all import Composition
>>> def build_composition(l):
...     return Composition(l)
>>> generator.set_builder(C, build_composition)

Now we can generate sage
`compositions <https://doc.sagemath.org/html/en/reference/combinat/sage/combinat/composition.html>`_: # noqa

>>> composition = generator.sample((10, 20))
>>> composition.parent()
Compositions of non-negative integers
>>> composition
[1, 2, 1, 2, 2, 2]
>>> ascii_art(composition)
       **
      **
     **
     *
    **
    *
"""



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

"""
Various oracle implementations for Boltzmann sampling.

Oracles are used to get (often approximate) values of generating functions.
Thanks to the symbolic method, functionnal equations can be derived from
grammar specifications. This module implements some mechanics to approximate
generating functions based on these equations.

Currently three oracles are implemented:

- :class:`OracleFromDict` is quite stupid ans wraps a dictionnary
  associating :py:class:`~usainboltz.grammar.Rule` to float value into an
  :class:`Oracle`
- :class:`OracleFromPaganini` use the oracle Pagagnini of [BBD2018]_
- :class:`OracleFromNewtonGF` use the oracle NewtonGF of [PSS2012]_
  (only when Sagemath and Maple installations are present)

The entry point of these algorithms is the function oracle which determines
the oracle to use given its inputs.

AUTHORS:
- Matthieu Dien (2019): initial version
- Martin Pépin (2019): initial version

.. rubric:: References

.. [BBD2018] Maciej Bendkowski, Olivier Bodini, Sergey Dovgal, 2018,
   Polynomial tuning of multiparametric combinatorial samplers, ANALCO
.. [PSS2012] Carine Pivoteau, Bruno Salvy, Michèle Soria, 2012
   Algorithms for Combinatorial Structures: Well-founded systems and
   Newton iterations, Journal of Combinatorial Theory
"""

import sys
from functools import partial, reduce

import paganini as pg
from cvxpy import SolverError

from usainboltz.grammar import (
    Atom,
    Cycle,
    Epsilon,
    Grammar,
    Marker,
    MSet,
    Product,
    RuleName,
    Seq,
    Set,
    Symbol,
    UCycle,
    Union,
)

__all__ = ["build_oracle", "OracleFromPaganini", "OracleFromDict"]


def build_oracle(sys, **kargs):
    """Build different oracles given different inputs

    EXAMPLES::
        >>> from usainboltz import *

        >>> leaf, z = Marker("leaf"), Atom()
        >>> B = RuleName("B")
        >>> g = Grammar({B: leaf + z * B * B})
        >>> build_oracle(g)
        OracleFromPaganini(B -> Union(leaf, Product(z, B, B)))

        >>> build_oracle({'B': 2, 'z': 1./4})
        OracleFromDict({'B': 2, 'z': 0.25})
    """

    if isinstance(sys, Grammar):
        return OracleFromPaganini(sys)
    elif isinstance(sys, dict):
        return OracleFromDict(sys)

    raise TypeError(
        "I do not know what to do with a system of type {}".format(type(sys))
    )


def _to_paganini_constraint(lower_size, upper_size):
    r"""Translates constraints over :py:class:`usainboltz.grammar.IteratedRule`
    into :py:class:`paganini.Constraint`"""
    if lower_size is not None and upper_size is not None:
        if lower_size == upper_size:
            return pg.eq(lower_size)
        else:
            raise ValueError("Paganini does not support mutliple constraints.")
    elif lower_size is not None:
        return pg.geq(lower_size)
    elif upper_size is not None:
        return pg.leq(upper_size)
    else:
        return None


def _to_paganini_rule(env, rule):
    r"""Recursively translates :py:class:`usainboltz.grammar.Rule`
    into :py:class:`paganini.Expr` using the environnement `env` to
    set the expectations of pagagnini `Variables`
    """
    if isinstance(rule, Epsilon):
        return pg.Expr()
    elif isinstance(rule, Symbol):  # means Atom, Marker or RuleName
        return env[rule]
    elif isinstance(rule, Union):
        return reduce(
            lambda x, y: x + y, map(partial(_to_paganini_rule, env), rule.args)
        )
    elif isinstance(rule, Product):
        return reduce(
            lambda x, y: x * y, map(partial(_to_paganini_rule, env), rule.args)
        )
    elif isinstance(rule, Seq):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Seq(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, Set):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Set(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, Cycle):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Cyc(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, MSet):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.MSet(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, UCycle):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.UCyc(_to_paganini_rule(env, rule.arg), constraint)
    #  elif isinstance(rule, PSet):
    #  raise NotImplementedError
    else:
        raise TypeError("first argument should be of type usainboltz.grammar.Rule")


def _to_paganini_spec(grammar, weights):
    r"""Convert a :py:class:`usainboltz.grammar.Grammar` with `weights`
    over symbols into a map :py:class:`usainboltz.grammar.Rule` ->
    :py:class:`paganini.Expr` and a :py:class:`paganini.Specification`

    Example:
        >>> from usainboltz import *
        >>> eps = Epsilon();  z = Atom(); B = RuleName()
        >>> grammar = Grammar({B: eps + z * B * B})

        >>> from usainboltz.oracle import _to_paganini_spec
        >>> env, spec = _to_paganini_spec(grammar, {z:1000})
        >>> env[z], env[B]
        (var1, var0)
        >>> spec
        var0 = 1  + var1^1 var0^2
    """

    spec = pg.Specification()
    env = dict()

    for rule in set(grammar.rules.keys()):
        if rule in weights:
            env[rule] = pg.Variable(weights[rule])
        else:
            env[rule] = pg.Variable()

    z = Atom()
    if z in weights:
        env[z] = pg.Variable(weights[z])
    else:
        env[z] = pg.Variable()

    for m in grammar.markers():
        if m in weights:
            env[m] = pg.Variable(weights[m])
        else:
            env[m] = pg.Expr()

    for rulename, rule in grammar.rules.items():
        spec.add(env[rulename], _to_paganini_rule(env, rule))

    return env, spec


class Oracle:
    pass


class OracleFromPaganini(Oracle):
    r"""Build an oracle using
    `paganini <https://github.com/maciej-bendkowski/paganini>`_ : a package
    developped by M. Bendkowski and S. Dovgal

    .. todo:: add the reference to the paper
    """

    def __init__(self, grammar, sanity_checks=True):
        r"""Build an oracle from a grammar

        Args:
          grammar (:py:class:`~usainboltz.grammar.Grammar`)

          sanity_checks (bool): enable sanity checks in paganini
            tuners (``True`` by default)

        Note:
          Set ``sanity_checks`` to ``False`` only if you know what you are doing !
        """
        super(OracleFromPaganini, self).__init__()
        self.grammar = grammar
        self.pg_method = pg.Method.STRICT if sanity_checks else pg.Method.FORCE

    def _env_to_val(self, env):
        r"""Wrapper for the output of :py:meth:`tuning`
        """
        values = {
            k: float(env[k].value if isinstance(env[k], pg.Variable) else 1.0)
            for k in env
        }
        values[Epsilon] = 1.0
        return values

    def tuning(self, rule, expectations=None, singular=False):
        r"""Run the oracle's algorithm

        Args:
          rule (:py:class:`~usainboltz.grammar.Rule`): the targeted
            symbol to tune w.r.t. ``expectations``

          expectations (dict): a mapping between grammar markers and
            their targeted expectations after tuning

          singular (bool): if ``True`` run the singular tuner
            (infinite expected size) else run a tuner targeting the
            ``expectations`` given

        Returns:
            values (dict): a mapping between grammar symbols and their
            weights in the Boltzmann model

        Note:
            If ``singular`` is ``True``, then ``expectations`` should
            be should be a dictionnary with grammar names (but the
            atom) as keys and ratios (between 0 and 1) as values.

            Else, ``expectations`` should be a dictionnary with keys as
            grammar names (but ``rule``) and the values should be
            integers.

        Examples:
           Use of the singular tuner for a binary tree grammar (size
           is the number of internal nodes):

           >>> from usainboltz import *
           >>> eps = Epsilon();  z = Atom(); B = RuleName()
           >>> grammar = Grammar({B: eps + z * B * B})
           >>> o = OracleFromPaganini(grammar)
           >>> values = o.tuning(B, singular=True)
           >>> abs(values[z]-0.25) < 1e-12
           True
           >>> abs(values[B]-2.) < 1e-12
           True

           As we expect ``z`` should be :math:`\frac14` (the
           singularity of :math:`\frac{1-\sqrt{1-4z}}{2}`)
        """

        if expectations is None:
            expectations = dict()

        if rule in expectations:
            raise ValueError("You must not set any constraint over {}".format(rule))

        # Check the values of expectations w.r.t. singular or not
        if singular:
            if any(x < 0 or x > 1 for x in expectations.values()):
                raise ValueError("Ratios must be between 0 and 1")
        elif any(x < 1 for x in expectations.values()):
            raise ValueError("The expected numbers must be greater than 1")

        env, spec = _to_paganini_spec(self.grammar, expectations)

        env[rule].set_expectation(None)
        for k, v in expectations.items():
            env[k].set_expectation(v)

        try:
            if singular:
                spec.run_singular_tuner(env[Atom()], method=self.pg_method)
            else:
                spec.run_tuner(env[rule], method=self.pg_method)
        except SolverError:
            raise ValueError("Your expected sizes can not be targeted")

        return self._env_to_val(env)

    def __repr__(self):
        return "OracleFromPaganini({})".format(self.grammar)


class OracleFromDict(Oracle):
    def __init__(self, dict_):
        super(OracleFromDict, self).__init__()
        if not (isinstance(dict_, dict)):
            raise TypeError("The argument should be a dictionnary")

        self.dict_ = dict_

    def tuning(self, *args, **kwargs):
        return self.dict_

    def __repr__(self):
        return "OracleFromDict({})".format(self.dict_)


def _to_combstruct_constraint(lower_size, upper_size):
    r"""Translates sizes constraint over
    :py:class:`usainboltz.grammar.IteratedRule` into `combstruct`
    constraints.

    Examples:
        >>> from usainboltz import * # doctest: +SKIP
        >>> z = Atom(); S = RuleName("S") # doctest: +SKIP
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, leq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, card <= 3)}'
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, geq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, 3 <= card)}'
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, eq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, card = 3)}'
    """

    x = (lower_size, upper_size)
    if x == (None, None):
        return ""
    elif x == (lower_size, None):
        return f", {lower_size} <= card"
    elif x == (None, upper_size):
        return f", card <= {upper_size}"
    else:
        if lower_size == upper_size:
            return f", card = {upper_size}"
        else:
            print(
                "WARNING: NewtonGF may not handle constraints "
                + "over lower and upper sizes",
                file=sys.stderr,
            )
            return f", card <= {upper_size}, {lower_size} <= card"


def _to_combstruct_rule(rule):
    r"""Recursively translates :py:class:`~usainboltz.grammar.Rule`
    into `combstruct[specification]`
    """

    if isinstance(rule, Epsilon):
        return "Epsilon"
    elif isinstance(rule, Atom):
        return "Atom"
    elif isinstance(rule, Marker):
        return "Epsilon"
    elif isinstance(rule, RuleName):
        return str(rule)
    elif isinstance(rule, Union):
        return "Union(" + ",".join(map(_to_combstruct_rule, rule.args)) + ")"
    elif isinstance(rule, Product):
        return "Prod(" + ",".join(map(_to_combstruct_rule, rule.args)) + ")"
    elif isinstance(rule, Seq):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Sequence({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, Set):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Set({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, Cycle):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Cycle({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, MSet):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Set({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, UCycle):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Cycle({_to_combstruct_rule(rule.arg)}{constraint})"
    #  elif isinstance(rule, PSet):
    #  raise NotImplementedError
    else:
        raise TypeError("first argument should be of type usainboltz.grammar.Rule")


def _to_combstruct_spec(grammar):
    spec = set()

    for marker in grammar.markers():
        spec.add(f"{marker}=Epsilon")

    for rulename, rule in grammar.rules.items():
        spec.add(f"{rulename} = {_to_combstruct_rule(rule)}")

    return "{" + ",".join(spec) + "}"


newtongf_installed = False

try:
    from sage.interfaces.maple import maple

    maple.load("combstruct")
    maple.load("NewtonGF")
    newtongf_installed = True
    __all__.append("OracleFromNewtonGF")
except ImportError:
    pass
except RuntimeError:
    pass


class OracleFromNewtonGF(Oracle):
    r"""Build an orcale using `NewtonGF
    <http://perso.ens-lyon.fr/bruno.salvy/software/the-newtongf-package/>`_

    Note:
        If you want to use OracleFormNewtonGF, please install the
        NewtonGF library for Maple into you Maple lib/ directory,
        or precise a lib/ directory to Maple i.e.
        maple('libname:=\"NewtonGF_path\",libname:')

"""

    def __init__(self, grammar):
        r"""Build an oracle from a grammar
        """

        if not newtongf_installed:
            raise ImportError(
                "You cannot use OracleFromNewtonGF because your "
                + "Sagemath, Maple or NewtonGF installation is not "
                + "consistent."
            )

        self.grammar = grammar
        self.combstruct_spec = _to_combstruct_spec(grammar)

    def tuning(self, rule, expectations=None, singular=False):
        r"""Run the oracle's algorithm

        Args:
          rule (:py:class:`~usainboltz.grammar.Rule`): the targeted
            symbol to tune w.r.t. ``expectations``

          expectations (dict): a mapping between grammar symbols and
            their targeted expectations after tuning

          singular (bool): if ``True`` run the singular tuner
            (infinite expected size) else run a tuner targeting the
            ``expectations`` given

        Returns:
            values (dict): a mapping between grammar symbols and their
            weights in the Boltzmann model

        Note:
            In the case of NewtonGF, only **univariate**
            specifications are supported, so:
            * if `expectations` is provided it must have a key corresponding
            to `~usainboltz.grammar.Atom()`,
            * the markers (see :py:class:`~usainboltz.grammar.Marker`)
            will not be handled.

        Examples:
           Use of the singular tuner for a binary tree grammar (size
           is the number of internal nodes):

           >>> from usainboltz import *
           >>> eps = Epsilon();  z = Atom(); B = RuleName()
           >>> grammar = Grammar({B: eps + z * B * B})
           >>> o = OracleFromNewtonGF(grammar) # doctest: +SKIP
           >>> values = o.tuning(B,singular=True) # doctest: +SKIP
           >>> values[z] # doctest: +SKIP
           0.25
           >>> values[B] # doctest: +SKIP
           1.999999988

           As we expect ``z`` should be :math:`\frac14` (the
           singularity of :math:`\frac{1-\sqrt{1-4z}}{2}`)

        """

        # To remove if we do the choice to autorize only one atom per grammar
        if len(self.grammar.markers()) > 0:
            print(
                "NewtonGF deals only with univariate specifications "
                + "and there are several markers in your one",
                file=sys.stderr,
            )

        labeled = "labeled" if self.grammar.labelled else "unlabeled"

        if not singular and (expectations is None or Atom() not in expectations):
            raise ValueError(
                "You must set an expectations over the size of the structure"
            )

        rho = 0
        if singular:
            rho = maple(f"Radius({self.combstruct_spec}, {labeled})")
        else:
            rho = maple(
                f"BoltzmannParameter({self.combstruct_spec}, {labeled}, {rule}, "
                + f"{expectations[Atom()]})"
            )

        oracle = maple(
            f"NumericalNewtonIteration({self.combstruct_spec}, {labeled})({rho})"
        )
        table = str().maketrans({"=": ":", "[": "{", "]": "}"})

        self.grammar_symbols = {
            str(symbol): symbol
            for symbol in self.grammar.markers() | set(self.grammar.rules)
        }

        values = eval(str(oracle).translate(table), globals(), self.grammar_symbols)
        values[Atom()] = float(rho)
        return values

    def __repr__(self):
        return "OracleFromNewtonGF({})".format(self.grammar)



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""

Index of examples

.. autosummary::

   expnums
   knr19
   mondrian

"""



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""Mondrian paintings generation.

Generate a Mondrian painting from a simple grammar.

>>> from usainboltz.examples.mondrian import *

>>> grammar
canvas -> Product(window, window, window, window)
window -> Union(Product(red, z),
                Product(blue, z),
                Product(yellow, z),
                Product(white, z),
                Product(black, z),
                Product(window, window, window, window))

The image is generated in svg format using the builders mechanism.

>>> sizes = {square: (20, 30), blue: (2, 4), red: (2, 4), yellow: (2, 4), black: (2, 4)}
>>> painting = generator.sample(sizes)
>>> print(painting)  # doctest: +ELLIPSIS
<svg width="1440" height="1440" xmlns="http://www.w3.org/2000/svg">
...

.. image:: ../_static/img/mondrian.svg
   :height: 480px
   :width: 480px
   :align: center
"""

from usainboltz import (
    Atom,
    Generator,
    Grammar,
    Marker,
    OracleFromPaganini,
    RuleName,
    union_builder,
)
from usainboltz.generator import seed as generator_seed

LENGTH = 5
STROKE = 4

red, blue, yellow, white, black = (
    Marker(color) for color in ("red", "blue", "yellow", "white", "black")
)
square = Atom()
window = RuleName("window")
canvas = RuleName("canvas")

grammar = Grammar(
    {
        canvas: window * window * window * window,
        window: red * square
        + blue * square
        + yellow * square
        + white * square
        + black * square
        + window * window * window * window,
    }
)

target_colors = {white: 0.6}

oracle = OracleFromPaganini(grammar, sanity_checks=False)
generator = Generator(
    grammar, canvas, singular=True, expectations=target_colors, oracle=oracle
)


def color_builder(t):
    color, _ = t
    return (
        f'<rect width="{LENGTH}" height="{LENGTH}" fill="{color}" '
        'vector-effect="non-scaling-stroke"/>',
        1,
    )


def pack_translate_scale(t):
    block, v, s = t
    return f'<g transform="translate({v[0]} {v[1]}), scale({s})">\n{block[0]}\n</g>'


def quad_builder(windows):
    sizes = [w[1] for w in windows]
    biggest = max(sizes)
    scale_factors = [biggest / x for x in sizes]
    translations = [
        (0, 0),
        (LENGTH * biggest, 0),
        (0, LENGTH * biggest),
        (LENGTH * biggest, LENGTH * biggest),
    ]
    return (
        "\n".join(map(pack_translate_scale, zip(windows, translations, scale_factors))),
        2 * biggest,
    )


window_builder = union_builder(
    color_builder,
    color_builder,
    color_builder,
    color_builder,
    color_builder,
    quad_builder,
)


def canvas_builder(windows):
    size = ((max([w[1] for w in windows])) * LENGTH * 2 + 40) * 4
    r = f'<svg width="{size}" height="{size}" xmlns="http://www.w3.org/2000/svg">'
    r += f'\n<g stroke="black" stroke-width="{STROKE}"'
    r += ' transform="scale(4) translate(20 20)">\n'
    r += quad_builder(windows)[0]
    r += "</g>\n</svg>"
    return r


generator.set_builder(window, window_builder)
generator.set_builder(canvas, canvas_builder)
generator_seed(0xDEADB017)

if __name__ == "__main__":
    print(
        generator.sample(
            {square: (20, 30), blue: (2, 4), red: (2, 4), yellow: (2, 4), black: (2, 4)}
        )
    )



# coding: utf-8

# cython: profile=True

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""Boltzmann generator for Context-free grammars.

This module provides functions for generating combinatorial objects (i.e.
objects described by a combinatorial specification, see
:mod:`usainboltz.grammar`) according to the Boltzmann
distribution.

Given an unlabelled combinatorial class A, the Boltzmann distribution of
parameter :math:`x` is such that an object of size n is drawn with the
probability :math:`\frac{x^n}{A(x)}` where :math:`A(x)` denotes the ordinary
generating function of A.  For labelled classes, this probability is set to
:math:`\frac{x^n}{n!A(x)}` where :math:`A(x)` denotes the exponential
generating function of A. See [DuFlLoSc04]_ for details.

By default, the objects produced by the generator are nested tuples of strings
(the atoms). For instance ``('z', ('z', 'epsilon', 'epsilon'), ('z', 'epsilon', 'epsilon'))`` is a
balanced binary tree with 3 internal nodes (z) and 4 leaves (e). To alter this
behaviour and generate other types of objects, you can specify a builder
function for each type of symbol in the grammar. The behaviour of the builders
is that they are applied in bottom up order to the structure "on the fly" during
generation. For instance, in order to generate Dyck words using the grammar for
binary trees, one can use a builder that return ``""`` for each leaf ``"(" +
left child + ")" + right child`` for each node. The builders will receive a
tuple for each product, a string for each atom and builders for unions should be
computed using the :func:`union_builder` helper. See the example below for the
case of Dyck words.

Examples:
    >>> from usainboltz import *

    Both binary trees and Dyck words can be generated from the following
    grammar

    >>> epsilon, z, B = Epsilon(), Atom(), RuleName("B")
    >>> grammar = Grammar({B: epsilon + z * B * B})
    >>> generator = Generator(grammar, B, singular=True)

    In order to build Dyck words, one can interpret epsilon as the empty word
    and nodes as the derivation: ``D -> ( D ) D``

    >>> def leaf_builder(_):
    ...     return ""

    >>> def node_builder(tuple):
    ...     _, left, right = tuple
    ...     return "(" + left + ")" + right

    >>> generator.set_builder(B, union_builder(leaf_builder, node_builder))
    >>> dyck_word = generator.sample((10, 20))
    >>> dyck_word  # random
    '((())(())())()()(())'
    >>> len(dyck_word) in range(20, 41)
    True

    If on the contrary we want to see trees as S-expressions, we can interpret
    epsilon as the leaf ``"leaf"`` and nodes as ``(node left_child
    right_child)``

    >>> def leaf_builder(_):
    ...     return "leaf"

    >>> def node_builder(tuple):
    ...     _, left, right = tuple
    ...     return "(node {} {})".format(left, right)

    >>> generator.set_builder(B, union_builder(leaf_builder, node_builder))
    >>> print(generator.sample((10, 20)))
    (node
      (node
        (node
          (node
            (node
              (node leaf leaf)
              (node
                leaf
                (node
                  (node leaf leaf)
                  (node
                    (node
                      (node
                        leaf
                        (node leaf leaf))
                      (node (node leaf leaf) (node leaf leaf)))
                    leaf))))
            leaf)
          (node leaf leaf))
        leaf)
      (node
        (node leaf leaf)
        (node leaf leaf)))

    Note that the builders mechanism can also be used to compute some
    statistics on the fly without explicitly building the whole structure. For
    instance the following example illustrates how to generate the **height**
    of a tree following the Boltzmann distribution without building the tree.

    >>> def leaf_height(_):
    ...     return 0

    >>> def node_height(tuple):
    ...     _, left, right = tuple
    ...     return 1 + max(left, right)

    >>> generator.set_builder(B, union_builder(leaf_height, node_height))
    >>> height = generator.sample((10, 20))
    >>> height  # random
    7

For labelled grammar, the procedure is the same. In this case, the default
builders also output the labelling, they do not only generate the structure.

Example:
    >>> z = Atom()
    >>> e, M = Epsilon(), RuleName()
    >>> g = Grammar({M: e + z * M + z * M * M}, labelled=True)
    >>> generator = Generator(g, M, singular=True)
    >>> generator.sample((5,10))  # random
    (2, (1, (4, (3, (5, 'epsilon'), 'epsilon'))), 'epsilon')

"""

import copy
from functools import reduce

from usainboltz.grammar import (
    Atom,
    Cycle,
    Epsilon,
    Marker,
    MSet,
    Product,
    RuleName,
    Seq,
    Set,
    Union,
    UCycle,
)
from usainboltz.oracle import build_oracle

from xoshiro cimport xoshiro
from libc.math cimport log, ceil, exp
from libc.stdint cimport uint64_t

cpdef seed(s):
    xoshiro.seed(s)

ctypedef enum instruction:
    # Symbols
    REF,
    ATOM,
    MARKER,
    EPSILON,
    # Grammar combinators
    UNION,
    PRODUCT,
    SEQ,
    SET,
    MSET,
    # Special instructions used by the random generators
    TUPLE,
    LIST,
    SET_PRODUCT,
    FUNCTION,
    WRAP_CHOICE,
    WRAPPED_CHOICE

# ------------------------------------------------------- #
# Grammar preprocessing
# ------------------------------------------------------- #

# For performance reasons, we use integers rather that strings to identify
# symbols during generation and we drop the python classes in favour of nested
# tuples. The following functions implement this transformation.
#
# All of this should remain hidden from the end user.

class Mapping:
    def __init__(self, grammar):
        # Rule names
        rule_name_to_id, id_to_rule_name = self._make_mapping(grammar.rules.keys())
        self.rule_name_to_id = rule_name_to_id
        self.id_to_rule_name = [id_to_rule_name[i] for i in range(len(id_to_rule_name))]
        # Markers
        marker_to_id, id_to_marker = self._make_mapping(grammar.markers())
        self.marker_to_id = marker_to_id
        self.id_to_marker = [id_to_marker[i] for i in range(len(id_to_marker))]

    def nb_rules(self):
        return len(self.id_to_rule_name)

    def nb_markers(self):
        return len(self.id_to_marker)

    @staticmethod
    def _make_mapping(iterable, start=0):
        x_to_id = {}
        id_to_x = {}
        for i, x in enumerate(iterable, start=start):
            x_to_id[x] = i
            id_to_x[i] = x
        return x_to_id, id_to_x

cdef float factorial(int n):
    cdef float res = 1.
    for i in range(2, n + 1):
        res *= float(i)
    return res


cdef _map_all_names_to_ids_expr(mapping, weights, expr):
    """Recursively transform an expression into a tuple of the form
    (RULE_TYPE, weight, args) where:
    - RULE_TYPE encodes the type of the expression [ATOM|REF|PRODUCT|...]
    - weight is the value of the generating function of this expression
    - args is auxilliary information (the name of an atom, the components of an
      union, ...)
    """
    # Symbols
    if isinstance(expr, RuleName):
        return (REF, weights[expr], mapping.rule_name_to_id[expr])
    elif isinstance(expr, Atom):
        return (ATOM, weights[expr], None)
    elif isinstance(expr, Marker):
        return (MARKER, weights[expr], (mapping.marker_to_id[expr], expr.name))
    elif isinstance(expr, Epsilon):
        return (EPSILON, 1., None)
    # Grammar combinators
    elif isinstance(expr, Union):
        args = tuple((_map_all_names_to_ids_expr(mapping, weights, arg) for arg in expr.args))
        total_weight = sum([w for (__, w, __) in args])
        return (UNION, total_weight, args)
    elif isinstance(expr, Product):
        args = tuple((_map_all_names_to_ids_expr(mapping, weights, arg) for arg in expr.args))
        total_weight = reduce(
            lambda x, y: x * y,
            [w for (__, w, __) in args],
            1.
        )
        return (PRODUCT, total_weight, args)
    elif isinstance(expr, Seq):
        rule = _map_all_names_to_ids_expr(mapping, weights, expr.arg)
        _, arg_weight, _ = rule
        full_series = 1. / (1. - arg_weight)
        weight = full_series
        if expr.lower_size is not None:
            weight *= arg_weight ** expr.lower_size
        if expr.upper_size is not None:
            weight -= arg_weight ** expr.upper_size * full_series
        return (SEQ, weight, (rule, expr.lower_size, expr.upper_size))
    elif isinstance(expr, Set):
        rule = _map_all_names_to_ids_expr(mapping, weights, expr.arg)
        _, arg_weight, _ = rule
        lower_size = expr.lower_size if expr.lower_size is not None else 0
        if expr.upper_size is not None:
            weight = sum(
                arg_weight ** i / factorial(i)
                for i in range(lower_size, expr.upper_size + 1)
            )
        else:
            weight = exp(arg_weight)
            weight -= sum(
                arg_weight ** i / factorial(i)
                for i in range(lower_size)
            )
        return (SET, weight, (rule, expr.lower_size, expr.upper_size))
    elif isinstance(expr, MSet):
        if expr.lower_size is not None or expr.upper_size is not None:
            raise NotImplementedError("Generator: size constraints on MSet")
        rule = _map_all_names_to_ids_expr(mapping, weights, expr.arg)
        weight = 0.  # this info should come from the oracle
        a_values = []  # this info shoud come from the oracle
        raise NotImplementedError("Generator: MSet weight")
        #  (MSET, weight, (rule, a_values))
    elif isinstance(expr, Cycle):
        raise NotImplementedError("Generator: Cycle")
    elif isinstance(expr, UCycle):
        raise NotImplementedError("Generator: UCycle")
    else:
        raise NotImplementedError("_map_all_names_to_ids({})".format(type(expr)))

cdef _map_all_names_to_ids_system(mapping, weights, rules):
    return [
        _map_all_names_to_ids_expr(mapping, weights, rules[mapping.id_to_rule_name[i]])
        for i in range(mapping.nb_rules())
    ]


# ------------------------------------------------------- #
# Generic builders
# ------------------------------------------------------- #

# When it chooses one possible derivation for a union rule, the random
# generator wraps the generated object in a tuple of the form ``(choice_id,
# object)`` so that the builder that will be called on this value has the
# information of which derivation was chosen. This helper function hides this
# machinery to the end user, allowing her to compose builders in a "high-level"
# manner.

def union_builder(*builders):
    """Factory for generating builders for union rules.

    The recommended way to write a builder for a union rule is to use this
    helper: write a auxilliary builder for each component of the union and
    compose them with ``union_builder``.

    Args:
        builders (List[function]): builder functions, one for each component of
            the union.

    Returns:
        function: a builder function for a disjoint union. The resulting
            function applies one of the functions passed as arguments to its
            input depending of the component of the union its argument belongs
            to.

    Examples:
        Assume the symbol ``D`` is defined by ``Union(A, B, C)``, then
        defining a builder for D looks like:

        >>> def build_A(args):
        ...     # Do something
        ...     pass

        >>> def build_B(args):
        ...     # Do something
        ...     pass

        >>> def build_C(args):
        ...     # Do something
        ...     pass

        >>> build_D = union_builder(build_A, build_B, build_C)

        For instance, for binary trees defined by ``B = Union(leaf, Product(z,
        B, B))``, this could be:

        >>> def build_leaf(_):
        ...     return BinaryTree()

        >>> def build_node(args):
        ...     z, left, right = args
        ...     return BinaryTree([left, right])

        >>> build_binarytree = union_builder(build_leaf, build_node)

    """
    def build(obj):
        index, content = obj
        builder = builders[index]
        return builder(content)
    return build

# The following functions generate the default builders for labelled and
# unlabelled structures. These builders produce nested tuples. Choices ids are
# omitted for readability.

cdef inline identity(x):
    return x

cdef inline ProductBuilder(builders):
    def build(terms):
        return tuple(builders[i](terms[i]) for i in range(len(terms)))
    return build

cdef inline SeqBuilder(builder):
    def build(terms):
        return [builder(term) for term in terms]
    return build

cdef make_default_builder(rule):
    """Generate the default builders for a rule.

    Args:
        rule (Rule): a grammar rule.

    Returns:
        function: a simple builder for the rule passed as argument that
            generate objects as nested tuples.
    """
    # Symbols
    if isinstance(rule, RuleName):
        return identity
    elif isinstance(rule, (Atom, Marker, Epsilon)):
        return identity
    # Grammar combinators
    elif isinstance(rule, Product):
        subbuilders = [make_default_builder(component) for component in rule.args]
        return ProductBuilder(subbuilders)
    elif isinstance(rule, Seq):
        subbuilder = make_default_builder(rule.arg)
        return SeqBuilder(subbuilder)
    elif isinstance(rule, Set):
        subbuilder = make_default_builder(rule.arg)
        # Same as Seq
        return SeqBuilder(subbuilder)
    elif isinstance(rule, Union):
        subbuilders = [make_default_builder(component) for component in rule.args]
        return union_builder(*subbuilders)
    else:
        raise NotImplementedError("make_default_builder({})".format(type(rule)))


# ------------------------------------------------------- #
# Random generation
# ------------------------------------------------------- #

cpdef int bounded_geometric(double param, lower_bound, upper_bound):
    cdef double r = xoshiro.rand_double()

    if lower_bound is None:
        lower_bound = 0

    bounding_term = 0
    if upper_bound is not None:
      k = upper_bound - lower_bound
      bounding_term = (1 - r) * (param ** (k + 1))

    geom = int(ceil(log(r + bounding_term)/log(param) - 1))
    return geom + lower_bound

cpdef int bounded_poisson(
        double total_weight,
        double param,
        lower_bound,
        upper_bound,
    ):
    cdef double r = xoshiro.rand_double() * total_weight
    cdef int i = lower_bound or 0
    cdef double s = 0.

    while s < r:
        s += param ** i / factorial(i)
        i += 1

    if upper_bound is not None:
        assert i <= upper_bound
    return (i - 1)

cpdef int mset_max_index(double total_weight, a_values):
    cdef double r = xoshiro.rand_double() * total_weight
    cdef double cumul = exp(a_values[0])
    cdef int j = 1

    while cumul < r:
        j += 1
        cumul *= exp(a_values[j] / j)

    return j


cdef _rand_perm(unsigned int n):
    """Draw an uniform random premutation of size ``n`` using the Fisher-Yates algorithm."""
    cdef unsigned int i, j
    cdef list p = list(range(1,n+1))

    for i in range(n-1,0,-1):
        j = xoshiro.rand_i64(i)
        p[i], p[j] = p[j], p[i]

    return p

# A free Boltzmann generator, but instead of generating a tree it only computes
# its sizes with respect to each of its atoms.

cdef c_simulate(first_rule, max_sizes, flat_rules):
    cdef sizes = [0 for _ in max_sizes]
    cdef list todo = [first_rule]
    cdef double r = 0.

    cdef instruction type
    cdef double weight

    while todo:
        type, weight, args = todo.pop()
        # Symbols
        if type == REF:
            symbol = args
            todo.append(flat_rules[symbol])
        elif type == ATOM:
            sizes[0] += 1
            if max_sizes[0] is not None and sizes[0] > max_sizes[0]:
                return sizes
        elif type == MARKER:
            symbol_id, _ = args
            symbol_id += 1
            sizes[symbol_id] += 1
            if max_sizes[symbol_id] is not None and sizes[symbol_id] > max_sizes[symbol_id]:
                return sizes
        elif type == EPSILON:
            pass
        # Grammar combinators
        elif type == UNION:
            r = xoshiro.rand_double() * weight
            for arg in args:
                __, arg_weight, __ = arg
                r -= arg_weight
                if r <= 0:
                    todo.append(arg)
                    break
        elif type == PRODUCT:
            todo += args[::-1]
        elif type == SEQ:
            arg, lower_size, upper_size = args
            _, arg_weight, _ = arg
            k = bounded_geometric(arg_weight, lower_size, upper_size)
            for __ in range(k):
                todo.append(arg)
        elif type == SET:
            arg, lower_size, upper_size = args
            _, arg_weight, _ = arg
            k = bounded_poisson(weight, arg_weight, lower_size, upper_size)
            for _ in range(k):
                todo.append(arg)
        elif type == MSET:
            arg, a_values = args
            k0 = mset_max_index(weight, a_values)
            for j in range(1, k0):
                param = a_values[j] / j
                n = bounded_poisson(exp(param), param, None, None)
                for _ in range(n):
                    # Generate j copies of Boltzmann(arg, z^j)
                    pass
            param = a_values[k0] / k0
            n = bounded_poisson(exp(param), param, 1, None)
            for _ in range(n):
                # Generate k0 copies of Boltzmann(arg, z^k0)
                pass
        else:
            raise NotImplementedError("c_simulate({})".format(type))
    return sizes

# Free Boltzmann sampler (actual generation) of *unlabelled* structures.
# Does not handle the labelling of labelled structures

cdef c_generate(first_rule, rules, builders, labels=None):
    cdef list generated = []
    cdef list todo = [first_rule]
    cdef double r = 0.

    while todo:
        type, weight, args = todo.pop()
        # Symbols
        if type == REF:
            symbol = args
            todo.append((FUNCTION, weight, symbol))
            todo.append(rules[symbol])

        elif type == ATOM:
            if labels is not None:
                generated.append(labels.pop())
            else:
                generated.append("z")
        elif type == EPSILON:
            generated.append("epsilon")
        elif type == MARKER:
            _, symbol_name = args
            generated.append(symbol_name)
        # Grammar combinators
        elif type == UNION:
            r = xoshiro.rand_double() * weight
            for i in range(len(args)):
                arg = args[i]
                __, arg_weight, __ = arg
                r -= arg_weight
                if r <= 0:
                    todo.append((WRAP_CHOICE, arg_weight, i))
                    todo.append(arg)
                    break
        elif type == PRODUCT:
            nargs = len(args)
            todo.append((TUPLE, weight, nargs))
            todo += args[::-1]
        elif type == SEQ:
            arg, lower_size, upper_size = args
            _, arg_weight, _ = arg
            k = bounded_geometric(arg_weight, lower_size, upper_size)
            todo.append((LIST, weight, k))
            for __ in range(k):
                todo.append(arg)
        elif type == SET:
            arg, lower_size, upper_size = args
            _, arg_weight, _ = arg
            k = bounded_poisson(weight, arg_weight, lower_size, upper_size)
            # store sets as lists
            todo.append((LIST, weight, k))
            for _ in range(k):
                todo.append(arg)
        # Internal instructions
        elif type == TUPLE:
            nargs = args
            t = tuple(generated[-nargs:])
            generated = generated[:-nargs]
            generated.append(t)
        elif type == LIST:
            nargs = args
            if nargs == 0:
                generated.append([])
            else:
                t = list(generated[-nargs:])
                generated = generated[:-nargs]
                generated.append(t)
        elif type == FUNCTION:
            func = builders[args]
            x = generated.pop()
            generated.append(func(x))
        elif type == WRAP_CHOICE:
            choice = generated.pop()
            choice_number = args
            generated.append((choice_number, choice))
        else:
            raise NotImplementedError("c_generate({})".format(type))

    obj, = generated
    return obj


# ------------------------------------------------------- #
# High level interface
# ------------------------------------------------------- #

cdef _in_sizes_windows(sizes, sizes_windows):
    for i, (size_min, size_max) in enumerate(sizes_windows):
        size = sizes[i]
        if size_min is not None and size < size_min:
            return False
        if size_max is not None and size > size_max:
            return False
    return True

cdef c_gen(first_rule, rules, sizes_windows, builders, labelling=False):
    """Search for a tree in a given size window."""
    cdef max_sizes = [size_max for (_, size_max) in sizes_windows]

    cdef uint64_t[4] backup_state

    # I'd need a do while
    xoshiro.get_state(backup_state)
    cdef sizes = c_simulate(first_rule, max_sizes, rules)
    while not _in_sizes_windows(sizes, sizes_windows):
        # save the random generator's state
        xoshiro.get_state(backup_state)
        sizes = c_simulate(first_rule, max_sizes, rules)

    labels = None
    if labelling:
        n = sizes[0]
        labels = _rand_perm(n)

    # Reset the random generator to the state it was just before the simulation
    xoshiro.set_state(backup_state)

    return c_generate(first_rule, rules, builders, labels=labels)

class GeneratorConfigError(Exception):
    pass

class Generator:
    """High level interface for Boltzmann samplers."""

    def __init__(
        self,
        grammar,
        rule_name=None,
        singular=None,
        expectations=None,
        oracle=None
    ):
        """Make a Generator out of a grammar.

        Args:
            grammar (Grammar): a combinatorial grammar

            rule_name (RuleName): the rule name of the symbol to generate.

            singular (bool): if set, do singular sampling (use the
                singularity of the generating functions as parameter).

            expectations (Dict[Symbol,number]): this is passed to the oracle to
                skew the distribution in order to target some specific values
                for the expected number of the specified symbols. See
                :py:meth:`OracleFromPaganini.tuning` for more details.

            oracle (Oracle): an oracle for computing the values of some
                generating functions. If not supplied, a default generic oracle
                that should work for most use-cases is automatically generated.

        Examples:

        >>> from usainboltz import *

        Some examples using the grammar for binary trees
        >>> z, B = Atom(), RuleName()
        >>> grammar = Grammar({B: Epsilon() + z * B * B})

        Typical use:
        >>> generator = Generator(grammar, B, singular=True)

        Since their is only one symbol in the grammar, the second argument can be
        omitted:
        >>> generator = Generator(grammar, singular=True)

        """
        self.grammar = copy.deepcopy(grammar)
        self.rule_name = self._guess_rule_name(rule_name)
        # Map all symbols in the grammar to an integer identifier.
        # Use arrays rather than dictionaries.
        self.mapping = Mapping(grammar)

        # If `singular` and the expectations are unspecified, do singular sampling
        # by default.
        if singular is None:
            singular = not(bool(expectations))

        # Set up the oracle.
        oracle = oracle or build_oracle(grammar)
        self.oracle_values = oracle.tuning(
            rule=Atom() if singular else rule_name,
            expectations=expectations,
            singular=singular
        )

        # Initialise the default builders.
        self.builders = [
            make_default_builder(self.grammar.rules[self.mapping.id_to_rule_name[id]])
            for id in range(self.mapping.nb_rules())
        ]

        # Convert the grammar into an array of tuples (for performance)
        self.flat_rules = _map_all_names_to_ids_system(
            self.mapping,
            self.oracle_values,
            self.grammar.rules
        )

    def _guess_rule_name(self, rule_name):
        if rule_name is not None:
            return rule_name
        rule_names = self.grammar.rules.keys()
        if len(rule_names) == 1:
            (rule_name, ) = rule_names
            return rule_name
        raise GeneratorConfigError(
            "I cannot guess which rule you want to generate. "
            "Specify a rule name using ``Generator(rule_name=...)``."
        )

    def _sanitize_windows(self, sizes_windows):
        if sizes_windows is None:
            sizes_windows = {}
        elif isinstance(sizes_windows, tuple):
            sizes_windows = {Atom(): sizes_windows}
        return [
            sizes_windows.get(Atom(), (None, None))
        ] + [
            sizes_windows.get(self.mapping.id_to_marker[i], (None, None))
            for i in range(self.mapping.nb_markers())
        ]


    def set_builder(self, non_terminal, func):
        """Set the builder for a non-terminal symbol.

        Args:
            non_terminal (RuleName): the name of the non-terminal symbol.

            func (function): the builder.
        """
        symbol_id = self.mapping.rule_name_to_id[non_terminal]
        self.builders[symbol_id] = func

    def get_builder(self, non_terminal):
        """Retrieve the current builder for a non-terminal symbol.

        Args:
            non_terminal (RuleName): the name of the non-terminal symbol.

        Returns:
            function: the current builder currently bound to the rule name
                passed as argument.
        """
        symbol_id = self.mapping.rule_name_to_id[non_terminal]
        return self.builders[symbol_id]

    def sample(self, sizes_windows):
        """Generate a term of the grammar given the prescribed sizes.

        Args:
            sizes_windows (Tuple[int,int]] | Dict[(Atom|Marker),Tuple[int,int]]):
                for atoms (resp. markers) for which a size window is
                specified, the generator will only generate objects whose
                number of those atoms (resp. markers) are in the
                specified size windows. For instance a
                ``generator.sample({z: (100, 200)})`` will only generate
                objects with a number of ``z`` belonging to the
                interval `[100, 200]`.
        """
        first_rule = (
            REF,
            self.oracle_values[self.rule_name],
            self.mapping.rule_name_to_id[self.rule_name]
        )
        return c_gen(
            first_rule,
            self.flat_rules,
            self._sanitize_windows(sizes_windows),
            self.builders,
            labelling=self.grammar.labelled,
        )



# coding: utf8

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

extensions = [
    Extension(
        "usainboltz.generator",
        sources=["src/usainboltz/generator.pyx", "src/xoshiro/xoshiro.c"],
        include_dirs=["src/xoshiro"],
        extra_compile_args=["-std=c99"]
    )
]

cython_aliases = None
include_dirs = None
exclude_packages = ["xoshiro"]

try:
    from sage import env

    cython_aliases = env.cython_aliases()
    include_dirs = env.sage_include_directories()
except ImportError:
    pass

setup(
    name="usainboltz",
    version="0.1b0",
    author="Matthieu Dien, Martin Pépin",
    author_email="kerl@wkerl.me",
    description="Fast Boltzmann random generators for SageMath",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/ParComb/usain-boltz",
    project_urls={
        "Bug Tracker": "https://gitlab.com/ParComb/usain-boltz/issues",
        "Documentation": "https://usain-boltz.readthedocs.io",
        "Source Code": "https://gitlab.com/ParComb/usain-boltz",
    },
    packages=find_packages("src", exclude=exclude_packages),
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, aliases=cython_aliases, language_level=3),
    install_requires=["paganini >= 1.3.2"],
    license="GPLv3+",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "Programming Language :: C",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    include_dirs=include_dirs,
    python_requires=">=3.6",
)



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""
Cayley trees
============

Cayley trees are non-ordered labeled trees. We can use SageMath
`LabelledRootedTree <http://doc.sagemath.org/html/en/reference/combinat/sage/combinat/rooted_tree.html>`_ # noqa
to build them.
We describe here how to write a grammar for Cayley trees in UsainBoltz
and how to generated sage objects.

The grammar of Cayley trees of size given by their number of vertices,
can be written as follows:

>>> from usainboltz import *
>>> from usainboltz.generator import seed
>>> seed(0xDEADBEEF)  # For reproducibility

>>> z = Atom()
>>> C = RuleName("C")
>>> grammar = Grammar({C: z * Set(C)}, labelled=True)
>>> grammar
C -> Product(z, Set(C))

To obtain a generator for this grammar, one must write:

>>> generator = Generator(grammar, C, singular=True)

But by default, UsainBoltz generates tuples:

>>> generator.sample((10, 20))
(9,
     [(4, [(2, [(6, [(10, [(3, [(11, [(5, [])]), (7, [(1, [(8, [])])])])])])])])])

Even if the default data structure generated use list it should be interpreted
as unordered.

Sage LabelledRootedTree can be obtained using the builders feature: we must tell the
generator how to build `C` objects:

>>> from sage.combinat.rooted_tree import LabelledRootedTree

A Cayley tree is a labeled root whose children are Cayley trees

>>> def build_tree(tupl):
...     z, sub_trees = tupl
...     return LabelledRootedTree(sub_trees, label=z)
>>> generator.set_builder(C, build_tree)

Now, the generator directly generates LabelledRootedTree objects:

>>> tree = generator.sample((10, 20))
>>> tree.parent()
Labelled rooted trees

>>> from sage.all import ascii_art
>>> ascii_art(tree)
          ____9______
         /          /   
        2         _4__
        |        / / / 
      __11__    3 6 13
     /     /   
    12   _5__  
        /   /  
       8   10  
           |   
           14  
          / /  
         1 7
"""



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""
Binary trees
============

Binary trees are plane trees with nodes of arity two and leaves of arity zero.
In sage binary trees are implemented in the
`binary_tree <https://doc.sagemath.org/html/en/reference/combinat/sage/combinat/binary_tree.html>`_ # noqa
module. We describe here how to write a grammar for binary trees in UsainBoltz
and how to generated sage objects.

The grammar of binary trees where only internal nodes are counted in the size
can be written as follows:

>>> from usainboltz import *
>>> from usainboltz.generator import seed
>>> seed(0xDEADBEEF)  # For reproducibility

>>> z, leave = Atom(), Epsilon()
>>> B = RuleName("B")
>>> grammar = Grammar({B: leave + z * B * B})
>>> grammar
B -> Union(epsilon, Product(z, B, B))

To obtain a generator for this grammar, one must write:

>>> generator = Generator(grammar, B, singular=True)

But by default, UsainBoltz generates tuples:

>>> generator.sample((10, 20))
('z',
     ('z',
      'epsilon',
      ('z',
       ('z', 'epsilon', 'epsilon'),
       ('z',
        ('z', 'epsilon', ('z', 'epsilon', 'epsilon')),
        ('z',
         'epsilon',
         ('z', 'epsilon', ('z', 'epsilon', ('z', 'epsilon', 'epsilon'))))))),
     'epsilon')

Sage binary trees can be obtained using the builders feature: we must tell the
generator how to build `B` objects:

>>> from sage.all import BinaryTree

- A leaf is `None`:

>>> def build_leaf(_):
...     return None

- A node is a call to `BinaryTree`:

>>> def build_node(tupl):
...     z, left, right = tupl
...     return BinaryTree([left, right])

- This two possible cases are combined using the :py:func:`~usainboltz.generator.union_builder` combinator:

>>> build_B = union_builder(build_leaf, build_node)
>>> generator.set_builder(B, build_B)

Now, the generator directly generates BinaryTree objects:

>>> tree = generator.sample((10, 20))
>>> tree.parent()
Binary trees

>>> tree
[[[[., .], [., .]], [., [[[., .], [., .]], .]]], .]

>>> from sage.all import ascii_art
>>> ascii_art(tree)
            o
           /
        __o__
       /     \
      o       o
     / \       \
    o   o       o
               /
              o
             / \
            o   o
"""



# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

"""Compatibility with Sage.

Within a sage environment, we use some sage functionnalities such as SageObject
in order to interact well with the sage ecosystem. Outside of sage, we implement
simplified versions of these tools to keep a similar behaviour.
"""

try:
    from sage.all import latex
    from sage.misc.fast_methods import Singleton

except ImportError:

    def latex(x):
        if hasattr(x, "_latex_"):
            return x._latex_()
        else:
            return str(x)

    class Singleton:
        _instance = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = object.__new__(cls)
            return cls._instance
